class StructuralAttributeBuilderFunctionsQueryBuilder extends GeneralFunctionsQueryBuilder {
  constructor(elements) {
    super(elements);
    this.elements = elements;

    document.querySelectorAll('[data-structural-attr-modal-action-button]').forEach(button => {
      button.addEventListener('click', () => {
        this.actionButtonInStrucAttrModalHandler(button.dataset.structuralAttrModalActionButton);
      });
    });
    document.querySelector('.ent-type-selection-action[data-ent-type="any"]').addEventListener('click', () => {
      this.submitQueryChipElement('start-empty-entity', 'Entity Start', '<ent>');
      this.submitQueryChipElement('end-entity', 'Entity End', '</ent>', null, true);
      this.elements.structuralAttrModal.close();
    });
    document.querySelector('.ent-type-selection-action[data-ent-type="english"]').addEventListener('change', (event) => {
      this.submitQueryChipElement('start-entity', `Entity Type=${event.target.value}`, `<ent_type="${event.target.value}">`, null, false, true);
      if (!this.elements.editingModusOn) {
        this.submitQueryChipElement('end-entity', 'Entity End', '</ent_type>', null, true);
      }
      this.elements.structuralAttrModal.close();
    });
    document.querySelector('.ent-type-selection-action[data-ent-type="german"]').addEventListener('change', (event) => {
      this.submitQueryChipElement('start-entity', `Entity Type=${event.target.value}`, `<ent_type="${event.target.value}">`, null, false, true); 
      if (!this.elements.editingModusOn) {
        this.submitQueryChipElement('end-entity', 'Entity End', '</ent_type>', null, true);
      }
      this.elements.structuralAttrModal.close();
    });
  }

  resetStructuralAttrModal() {
    this.resetMaterializeSelection([this.elements.englishEntTypeSelection, this.elements.germanEntTypeSelection]);
    this.resetMaterializeSelection([this.elements.textAnnotationSelection], 'address');
    this.elements.textAnnotationInput.value = '';

    this.toggleClass(['entity-builder', 'text-annotation-builder'], 'hide', 'add');
    this.toggleEditingAreaStructureAttrModal('remove');
    this.elements.editingModusOn = false;
    this.elements.editedQueryChipElementIndex = undefined;
  }

  toggleEditingAreaStructureAttrModal(action) {
    // If the user edits a query chip element, the corresponding editing area is displayed and the other areas are hidden or disabled.
    this.toggleClass(['sentence-button', 'entity-button', 'text-annotation-button', 'any-type-entity-button'], 'disabled', action);
  }

  actionButtonInStrucAttrModalHandler(action) {
    switch (action) {
      case 'sentence':
        this.submitQueryChipElement('start-sentence', 'Sentence Start', '<s>');
        this.submitQueryChipElement('end-sentence', 'Sentence End', '</s>', null, true);
        this.elements.structuralAttrModal.close();
        break;
      case 'entity':
        this.toggleClass(['entity-builder'], 'hide', 'toggle');
        this.toggleClass(['text-annotation-builder'], 'hide', 'add');
        break;
      case 'meta-data':
        this.toggleClass(['text-annotation-builder'], 'hide', 'toggle');
        this.toggleClass(['entity-builder'], 'hide', 'add');
        break;
      default:
        break;
    }
  }

  textAnnotationSubmitHandler() {
    let noValueMetadataMessage = document.querySelector('#corpus-analysis-concordance-no-value-metadata-message');
    let textAnnotationSubmit = document.querySelector('#corpus-analysis-concordance-text-annotation-submit');
    let textAnnotationInput = document.querySelector('#corpus-analysis-concordance-text-annotation-input');
    let textAnnotationOptions = document.querySelector('#corpus-analysis-concordance-text-annotation-options');

    if (textAnnotationInput.value === '') {
      textAnnotationSubmit.classList.add('red');
      noValueMetadataMessage.classList.remove('hide');
      setTimeout(() => {
        textAnnotationSubmit.classList.remove('red');
      }, 500);
      setTimeout(() => {
        noValueMetadataMessage.classList.add('hide');
      }, 3000);
    } else {
      let queryText = `:: match.text_${textAnnotationOptions.value}="${textAnnotationInput.value}"`;
      this.submitQueryChipElement('text-annotation', `${textAnnotationOptions.value}=${textAnnotationInput.value}`, queryText, null, false, true);
      this.elements.structuralAttrModal.close();
    }
  }
}
